/***********************************************************************
*
* ush.c
*
* uvscan smtp helper
* This helper applications allows uvscan to be called as a filter
* by ProxSMTP by Nate Nielsen. uvscan is the commandline antivirus
* scanner for Linux by McAfee and is licensed per server.
* For more information:
* http://www.ward.nu/computer/tools
* http://memberwebs.com/nielsen/software/proxsmtp/
* http://www.sng.ecs.soton.ac.uk/cgi-bin/faq?_recurse=1&file=15#file_213
*
* Copyright (C) 2005 Ward van Wanrooij <ward@ward.nu>
*
* This plugin may be distributed according to the terms of the GNU
* General Public License, version 2 or (at your option) any later version.
*
***********************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>

#define PIPE_READ 0
#define PIPE_WRITE 1
#define EXIT_TRANSFAILURE 2

void internal_error(const char* buf, int err);
int copyfile(const char* infile, const char* outfile);

int main(int argc, char* argv[]) {
	int pipefd[2];
	int status;
	pid_t pid;
	char buf[1024];
	char* sdesc;
	char* par_uvscan=NULL;
	char* par_uvconf=NULL;
	char* par_infstore=NULL;

	if (argc<3) {
		fprintf(stderr, "ush 20050109 - uvscan smtp helper\nSyntax: %s /path/to/uvscan /path/to/uvscan.conf [/path/to/infected]\n\n", argv[0]);
		internal_error("invalid number of arguments", argc);
	}
	par_uvscan=argv[1];
	par_uvconf=argv[2];
	if (argc==4) {
		par_infstore=argv[3];
	}
	if (getenv("EMAIL")==NULL) internal_error("EMAIL unspecified", 0);

	if (pipe(pipefd)) internal_error("unable to create pipe", errno);

	pid=fork();
	if (pid==-1) {
		internal_error("unable to fork", errno);
	} if (pid==0) { //child
		close(pipefd[PIPE_READ]);
		close(STDOUT_FILENO);
		if (dup2(pipefd[PIPE_WRITE], STDOUT_FILENO)==-1) internal_error("unable to dup2 pipe_write", errno);
		execl(par_uvscan, par_uvscan, "--config", par_uvconf, getenv("EMAIL"), NULL);
		return(EXIT_TRANSFAILURE);
	} else { //parent
		int cnt=0;
		int n;

		close(pipefd[PIPE_WRITE]);
		close(STDIN_FILENO);
		if (dup2(pipefd[PIPE_READ], STDIN_FILENO)==-1) internal_error("unable to dup2 pipe_read", errno);
		memset(buf, 0, sizeof(buf));
		while (cnt<sizeof(buf)-1) {
			n=read(STDIN_FILENO, buf+cnt, sizeof(buf)-cnt-1);
			if (n<=0) break;
			cnt+=n;
		}
		close(pipefd[PIPE_READ]);
		waitpid(pid, &status, 0);
	}

	switch (WIFEXITED(status) ? WEXITSTATUS(status) : 6) {
		case 0:
			//clean
			return(EXIT_SUCCESS);
		case 12:
		case 13:
		case 19:
			//infected
			sdesc=index(buf, 10);
			if (sdesc!=NULL) {
	                        char* edesc=0;
				sdesc++;
				edesc=index(sdesc, 10);
				if (edesc!=NULL) *edesc=0;
				while (sdesc<buf+sizeof(buf)) {
					if (*sdesc==32) sdesc++;
					else break;
				}
			} else {
				sdesc=buf;
			}
			char filename[1024];
			memset(filename, 0, sizeof(filename));
			if (par_infstore!=NULL) {
				snprintf(filename, sizeof(filename)-1, "%s/%d-%d.msg", par_infstore, time(), (int)getpid());
				copyfile(getenv("EMAIL"), filename);
				openlog("ush", 0, LOG_MAIL);
				syslog(LOG_INFO, "Infected message (return=%d) from pid %d stored as %s", WEXITSTATUS(status), (int)getpid(), filename);
				closelog();
			}
                        fprintf(stderr, "550 Contaminated: %s (%d)\n", sdesc, (int)getpid());
			return(EXIT_FAILURE);
		case 2:
		case 6:
		case 8:
		case 15:
		case 102:
		default:
			//error
			internal_error("exitcode", WEXITSTATUS(status));
	}
}

void internal_error(const char* buf, int err) {
	fprintf(stderr, "451 Internal error: %s (%d)\n", buf, err);
	exit(EXIT_TRANSFAILURE);
}

int copyfile(const char* infile, const char* outfile) {
	int infd;
	int outfd;
	int incnt;
	int outcnt;
	char buf[8192];

	infd=open(infile, O_RDONLY);
	if (infd==-1) internal_error("error opening infile", errno);
	outfd=open(outfile, O_WRONLY|O_CREAT, 0644);
	if (outfd==-1) internal_error("error opening outfile", errno);
	while (1) {
		incnt=read(infd, buf, sizeof(buf));
		if (incnt==0) break;
		if (incnt<0) internal_error("error reading infile", errno);
		outcnt=write(outfd, buf, incnt);
		if (incnt!=outcnt) internal_error("error writing outfile", errno);
	}
	close(outfd);
	close(infd);
	return(0);
}

